// Find all solutions to a^2 + b^2 = c^2 + d^2 where a, b, c, d in [0, 1000].
// Taken from G. Laakmann McDowell: "Cracking the Coding Interview", Ch. VII Technical Questions

#include <iostream>
#include <chrono>
#include <string>

#include <cmath>          // used in function better
#include <unordered_map>  // used in functions good and better
#include <utility>        // std::pair // used in functions good and better
#include <vector>         // used in functions good and better

/* 
 * Four-fold for-loop 
 */
void naive (bool print) {
  for (unsigned a = 0; a <= 1000; ++a) 
    for (unsigned b = 0; b <= 1000; ++b)
      for (unsigned c = 0; c <= 1000; ++c)
        for (unsigned d = 0; d <= 1000; ++d)
          if (a*a + b*b == c*c + d*d)
            if (print)
              std::cout 
                << "a: " << a 
                << "  |  b: " << b
                << "  |  c: " << c 
                << "  |  d: " << d
                << std::endl;
}


/* 
 * Three-fold for-loop and solve equation for variable d 
 */
void better (bool print) {
  for (unsigned a = 0; a <= 1000; ++a) {
    for (unsigned b = 0; b <= 1000; ++b) {
      for (unsigned c = 0; c <= 1000; ++c) {
        double d_double = std::sqrt(a*a + b*b - c*c);
        if (d_double >= 0 && d_double <= 1000) {
          unsigned d = (unsigned) d_double;
          if (a*a + b*b == c*c + d*d) {
            if (print) {
            std::cout << "a: " << a << "  |  b: " << b
                     << "  |  c: " << c << "  |  d: " << d
                     << std::endl;;
            }
          }
        }
      }
    }
  }
}


/* 
 * Build a hash map that maps the result of a calculation to the parameters 
 */
void good (bool print) {
  // pair of two numbers = one side of the eq.
  using pair_t = std::pair<unsigned, unsigned>;  
  // list of pairs
  using list_t = std::vector<pair_t>;

  // map that maps the calculation result (number) to list of pairs of numbers
  std::unordered_map<unsigned, list_t> map;

  // Build the map
  for (unsigned a = 0; a <= 1000; ++a) {
    for (unsigned b = 0; b <= 1000; ++b) {
      unsigned lhs = a*a + b*b;
      // emplace_back takes arguments of an object constructor, 
      // constructs that object using the resp. constructor, 
      // and stores it in container
      map[lhs].emplace_back(a, b);
    }
  }

  // Probe the map
  for (unsigned c = 0; c <= 1000; ++c) {
    for (unsigned d = 0; d <= 1000; ++d) {
      unsigned rhs = c*c + d*d;  // lookup key
      const list_t& list = map[rhs];  // list of parameter pairs (c, d) for which c*c + d*d == key
      for (const pair_t& lhs : list) {  // Range-based for loop ("for each element")
        if (print) {
          std::cout << "a: " << lhs.first << "  |  b: " << lhs.second
                    << "c: " << c         << "  |  d: " << d
                    << std::endl;
        }
      }
    }
  }
}


/*
 * Hash map approach like above, but no unnecessary re-calculation
 */
void smart (bool print) {
  using pair_t = std::pair<unsigned, unsigned>;
  using list_t = std::vector<pair_t>;

  std::unordered_map<unsigned, list_t> map;

  for (unsigned a = 0; a <= 1000; ++a) {
    for (unsigned b = 0; b <= 1000; ++b) {
      unsigned lhs = a*a + b*b;
      map[lhs].emplace_back(a, b);
    }
  }

  // auto is of type: std::pair<const unsigned, std::vector<std::pair<unsigned, unsigned>>>
  for(const auto& bucket : map) {  // Go through each entry in the map
    for (const pair_t& lhs : bucket.second) {  // Go through each pair of pairs...
      for (const auto& rhs : bucket.second) {  // ...
        if (print) {
          std::cout << "a: " << lhs.first << "  |  b: " << lhs.second
                    << "  |  c: " << rhs.first << "  |  d: " << rhs.second
                    << std::endl;
        }
      }
    }
  }
}


/*
 * Like smart, but less loop iterations due to commutativity of '+'
 */
void awesome (bool print) {
  using pair_t = std::pair<unsigned, unsigned>;
  using list_t = std::vector<pair_t>;

  std::unordered_map<unsigned, list_t> map;

  for (unsigned a = 0; a <= 1000; ++a) {
    for (unsigned b = a; b <= 1000; ++b) {
      unsigned lhs = a*a + b*b;
      map[lhs].emplace_back(a, b);
      map[lhs].emplace_back(b, a);
    }
  }

  // auto is of type: std::pair<const unsigned, std::vector<std::pair<unsigned, unsigned>>>
  for(const auto& bucket : map) {  // Go through each entry in the map
    for (const pair_t& lhs : bucket.second) {  // Go through each pair of pairs...
      for (const auto& rhs : bucket.second) {  // ...
        if (print) {
          std::cout << "a: " << lhs.first << "  |  b: " << lhs.second
                    << "  |  c: " << rhs.first << "  |  d: " << rhs.second
                    << std::endl;
        }
      }
    }
  }
}


int main (int argc, char* argv []) {
  if (argc != 3) {
    std::cout << "usage: " << argv[0] << " <functionname> <print={0,1}" << std::endl;
    return -1;
  } 

  std::string functionname (argv[1]);  // get command line argument
  
  using func_t = void (*) (bool);      // function pointer typedef
  func_t func;
  
  if (functionname == "naive") {
    func = naive;
  } else if (functionname == "better") {
    func = better;
  } else if (functionname == "good") {
    func = good;
  } else if (functionname == "smart") {
    func = smart;
  } else if (functionname == "awesome") {
    func = awesome;
  }  else {
    std::cout << "unknown functionname" << std::endl;
    return -1;
  }

  std::string printOption (argv[2]);
  bool print;
  if (printOption == "0") {
    print = false;
  } else if (printOption == "1") {
    print = true;
  } else {
    std::cout << "unsupported print option" << std::endl;
    return -1;
  }

  // Execute selected function and measure time
  std::chrono::high_resolution_clock::time_point startpoint = std::chrono::high_resolution_clock::now();
  func(print);
  std::chrono::high_resolution_clock::time_point endpoint = std::chrono::high_resolution_clock::now();
  
  unsigned runtimeMs = std::chrono::duration_cast<std::chrono::milliseconds> (endpoint - startpoint).count();
  
  std::cout << "runtime of " << functionname << " in milliseconds: " << runtimeMs << std::endl;
}
