/*
 * Tests for the templated stack 
 * Author: Daniel Flachs
 */

#include <cstddef>
#include <random>
#include <iomanip>
#include <chrono>

#include "stack.hh"


inline void printHeader() {
    std::cout << std::string(50, '-') << std::endl;
    std::cout 
        << std::setw(8) << "i"
        << std::setw(8) << "size"
        << std::setw(8) << "top"
        << std::setw(8) << "op"
        << std::setw(8) << "val-1"
        << std::setw(8) << "Status"
        << std::endl;
    std::cout << std::string(50, '-') << std::endl;
}

inline std::string opToStr(unsigned op) {
    switch(op) {
        case 1: 
            return "push";
        case 2: 
            return "pop";
        case 3: 
            return "top";
        case 4: 
            return "empty";
        default: 
            return "err";
    }
}


int testStack(const size_t noTestOps, const bool debug = false) {
    // test Stack with noTestOps many random Stack operation

    Stack<unsigned> s (1);

    std::default_random_engine generator;
    std::uniform_int_distribution<int> distribution(1,4);

    unsigned val = 0;

    unsigned noPush = 0;
    unsigned noPop = 0;
    unsigned noTop = 0;
    unsigned noEmpty = 0;

    for (unsigned i = 0; i < noTestOps; ++i) {
        unsigned op = distribution(generator);
        if (s.size() == 0) { op = 1; }
        switch(op) {
            case 1: 
                s.push(val++); 
                ++noPush;
                break;
            case 2: 
                try { s.pop(); } catch(std::exception e) {} 
                ++noPop;
                --val;
                break;
            case 3: 
                try { s.top(); } catch(std::exception e) {} 
                ++noTop;
                break;
            case 4: 
                s.empty(); 
                ++noEmpty;
                break;
            default: 
                std::cout << "error state" << std::endl; 
                return 1;
        }

        if (debug) {
            if (i % 30 == 0) { printHeader(); }
            
            std::cout 
                << std::setw(8) << i 
                << std::setw(8) << s.size()
                << std::setw(8) << ((!s.empty()) ? s.top() : 0)
                << std::setw(8) << opToStr(op)
                << std::setw(8) << (val-1)
                << std::setw(8) << (!s.empty() && s.top() != (val-1) ? "ERR" : "ok")
                << std::endl;
        }
    }

    if (debug) { std::cout << std::endl; }


    size_t colw = 25;
    const std::string sepline(2*colw, '-');
    const std::string blankline(2*colw, ' ');

    std::cout
        << "| " << sepline  << " |" << std::endl
        << "| TEST STATISTICS                                    |" << std::endl
        << "| " << sepline  << " |" << std::endl
        << "| " << blankline  << " |" << std::endl
        << std::setfill('_')
        << "| " << std::left << std::setw(colw) << "push" << std::right << std::setw(colw) << noPush  << " |" << std::endl
        << "| " << std::left << std::setw(colw) << "pop" << std::right << std::setw(colw) << noPop   << " |" << std::endl
        << "| " << std::left << std::setw(colw) << "top" << std::right << std::setw(colw) << noTop   << " |" << std::endl
        << "| " << std::left << std::setw(colw) << "empty" << std::right << std::setw(colw) << noEmpty << " |" << std::endl
        << "| " << std::left << std::setw(colw) << "total" << std::right << std::setw(colw) 
               << noPush + noPop + noTop + noEmpty << " |" << std::endl
        << "| " << blankline  << " |" << std::endl
        << "| " << std::left << std::setw(colw) << "Stack reallocations" 
            << std::right << std::setw(colw) << s.reallocs()  << " |" << std::endl
        << "| " << std::left << std::setw(colw) << "Time per reallocation" 
            << std::right << std::setw(colw) << std::setprecision(2) << std::fixed 
            << s.timePerRealloc()  << " |" << std::endl
        << "| " << std::left << std::setw(colw) << "Stack size" 
            << std::right << std::setw(colw) << s.size()  << " |" << std::endl
        << "| " << std::left << std::setw(colw) << "Stack capacity" 
            << std::right << std::setw(colw) << s.capacity()  << " |" << std::endl
        << "| " << blankline  << " |" << std::endl
        << "| " << sepline  << " |" << std::endl;

    return 0;
}


int main() {

    size_t numOps = 1e6;

    time_point start = std::chrono::high_resolution_clock::now();
    testStack(numOps);
    time_point stop  = std::chrono::high_resolution_clock::now();

    std::cout << std::endl;
    std::cout 
        << "Duration per stack operation: " 
        << (nanosec(stop - start).count()) / numOps 
        << " ns"
        << std::endl;
    
    
    // Stack<int> s(1);

    // std::cout << "Stack size: " << s.size() << std::endl;
    // 
    // s.push(12);
    // s.push(5);
    // s.push(100);

    // std::cout << "Stack size: " << s.size() << std::endl;

    // // Not possible: 
    // // error: binding value of type 'const int' to reference to type 'int' 
    // //        drops 'const' qualifier  
    // // int &y = s.top();
    // // x = 3;

    // const int &x = s.top();
    // std::cout << "s.top():    " << x << std::endl;

    // std::cout << "Stack size: " << s.size() << std::endl;
    // std::cout << "s.pop():    " << s.pop() << std::endl;
    // std::cout << "Stack size: " << s.size() << std::endl;
    // std::cout << "s.pop():    " << s.pop() << std::endl;
    // std::cout << "Stack size: " << s.size() << std::endl;
    // std::cout << "s.pop():    " << s.pop() << std::endl;
    // std::cout << "Stack size: " << s.size() << std::endl;
    // //std::cout << s.pop() << std::endl;
}
