/* 
 * Implementation of a templated stack (LIFO container)
 * Author: Daniel Flachs (dflachs@mail.uni-mannheim.de)
 */

#ifndef STACK2_HH
#define STACK2_HH

#include <cstddef>
#include <iostream>
#include <cassert>
#include <cstring>
#include <algorithm>
#include <chrono>

#define DEFAULT_SIZE 1

using time_point = std::chrono::high_resolution_clock::time_point;
using nanosec  = std::chrono::duration<long double, std::nano>;
using millisec = std::chrono::duration<long double, std::milli>;


template <typename T>
class Stack {
    public:
        /* Constructors */
        Stack(const size_t aCapacity, const bool aTrace = false);
        Stack();

        /* Destructor */
        ~Stack() { delete [] _data; }

        /* Container Operations */
        T pop();
        const T& top();
        void push(const T& aItem);

        /* Getters */
        inline size_t      capacity()       const { return _capacity; }
        inline size_t      size()           const { return _size; }
        inline bool        empty()          const { return 0 == _size; }
        inline bool        full()           const { return _capacity == _size; }
        inline size_t      reallocs()       const { return _reallocs; }
        inline long double timePerRealloc() const { return (_reallocTime.count() / _reallocs); }

    private:
        /* Data Members */
        T* _data;        
        size_t _capacity;
        size_t _size;

        bool _trace;           // Print debug messages
        size_t _reallocs;      // Counter for reallocations/copies
        nanosec _reallocTime;  // Total time needed for copies

        /* Methods */
        void expand();
};


/* -- Implementation starts here --  */

template <typename T>
Stack<T>::Stack(const size_t aCapacity, const bool aTrace) 
    : _data(nullptr), 
      _capacity(aCapacity), 
      _size(0),
      _trace(aTrace),
      _reallocs(0) {
    
    _data = new T[_capacity];
} 

template <typename T>
Stack<T>::Stack() 
    : Stack(DEFAULT_SIZE) {}

// push
template <typename T>
void Stack<T>::push(const T& aItem) {
    assert(_capacity >= _size);
    if (full()) {
        expand();
    }
    if (_trace) { std::cout << "push()" << std::endl; }
    _data[_size] = aItem;
    ++_size;
}

// pop
template <typename T>
T Stack<T>::pop() {
    if (!empty()) {
        --_size;
        return _data[_size];
    } else {
        throw std::range_error("Stack size is 0. Nothing to return.");
    }
}

// top
template <typename T>
const T& Stack<T>::top() {
    if (!empty()) {
        return _data[_size - 1];
    } else {
        throw std::range_error("Stack size is 0. Nothing to return.");
    }
}

// expand
template <typename T>
void Stack<T>::expand() {
    if (_trace) { 
        std::cout 
            << "expand(): capacity " 
            << _capacity << " -> " << (2*_capacity) 
            << std::endl;
    }
    
    _capacity *= 2;
    T* olddata = _data;        // Save pointer to old data
    _data = new T[_capacity];  // Allocate new memory

    time_point start = std::chrono::high_resolution_clock::now();
    // Copy data from old to new location
    //for (size_t i = 0; i < _size; ++i) {
    //    _data[i] = olddata[i];
    //}
    //memcpy(_data, olddata, _size * sizeof(T)); 
    std::copy(olddata, olddata+_size, _data);
    time_point stop  = std::chrono::high_resolution_clock::now();
    _reallocTime += nanosec(stop - start);

    delete [] olddata;         // Free old memory
    ++_reallocs;
}

#endif
